#include "mainwindow.h"
#include <QMidiInput>
#include <QMidiSystemNotifier>
#include <QDebug>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>

MainWindow::MainWindow(QWidget* parent)
    : QMainWindow(parent) {
    QWidget* mainWidget = new QWidget(this);
    setCentralWidget(mainWidget);
    QVBoxLayout *mainLayout = new QVBoxLayout(mainWidget);
    mainWidget->setLayout(mainLayout);

    // comboboxes
    QHBoxLayout *conLayout = new QHBoxLayout(mainWidget);
    mainLayout->addLayout(conLayout);

    mInputDevices = new QComboBox(this);
    mOutputDevices = new QComboBox(this);
    QLabel *forwardText = new QLabel(QString::fromLatin1("-->"), this);
    conLayout->addWidget(mInputDevices, 2);
    conLayout->addWidget(forwardText, 1);
    conLayout->addWidget(mOutputDevices, 2);

    connect(mInputDevices, SIGNAL(currentIndexChanged(int)), this, SLOT(inputIndexChanged(int)));
    connect(mOutputDevices, SIGNAL(currentIndexChanged(int)), this, SLOT(outputIndexChanged(int)));

    // monitor
    mConsoleOutput = new QPlainTextEdit(this);
    mainLayout->addWidget(mConsoleOutput);

    mConsoleOutput->setReadOnly(true);
    QFont font("Courier New");
    mConsoleOutput->setFont(font);

    addLine("MIDI Device Monitor");
    addLine("===================");
    addLine("");

    auto notifier = new QMidiSystemNotifier(this);
    connect(notifier, &QMidiSystemNotifier::inputDeviceAttached, this, &MainWindow::inputDeviceAttached);
    connect(notifier, &QMidiSystemNotifier::outputDeviceAttached, this, &MainWindow::outputDeviceAttached);
    connect(notifier, &QMidiSystemNotifier::inputDeviceDetached, this, &MainWindow::inputDeviceDetached);
    connect(notifier, &QMidiSystemNotifier::outputDeviceDetached, this, &MainWindow::outputDeviceDetached);

    connect(notifier, &QMidiSystemNotifier::inputDeviceCreated, this, &MainWindow::inputDeviceCreated);
    connect(notifier, &QMidiSystemNotifier::outputDeviceCreated, this, &MainWindow::outputDeviceCreated);
    connect(notifier, &QMidiSystemNotifier::inputDeviceDeleted, this, &MainWindow::inputDeviceDeleted);
    connect(notifier, &QMidiSystemNotifier::outputDeviceDeleted, this, &MainWindow::outputDeviceDeleted);
}

MainWindow::~MainWindow() {
}

void MainWindow::addLine(QString line) {
    //QTextCursor prev_cursor = mConsoleOutput->textCursor();
    mConsoleOutput->moveCursor(QTextCursor::End);
    mConsoleOutput->insertPlainText(line + "\n");
    mConsoleOutput->moveCursor(QTextCursor::End);
    //mConsoleOutput->setTextCursor(prev_cursor);
}

void MainWindow::inputDeviceAttached(const QMidiDeviceInfo info) {
    addLine(QString("Input device attached: %1").arg(info.deviceName()));
    mInputDevices->addItem(info.deviceName(), QVariant::fromValue(info));
}

void MainWindow::outputDeviceAttached(const QMidiDeviceInfo info) {
    addLine(QString("Output device attached: %1").arg(info.deviceName()));
    mOutputDevices->addItem(info.deviceName(), QVariant::fromValue(info));
}

void MainWindow::inputDeviceDetached(const QMidiDeviceInfo info) {
    addLine(QString("Input device detached: %1").arg(info.deviceName()));
    mInputDevices->removeItem(mInputDevices->findData(QVariant::fromValue(info)));
}

void MainWindow::outputDeviceDetached(const QMidiDeviceInfo info) {
    addLine(QString("Output device detached: %1").arg(info.deviceName()));
    mOutputDevices->removeItem(mOutputDevices->findData(QVariant::fromValue(info)));
}

void MainWindow::inputDeviceCreated(const QMidiInput* device) {
    addLine(QString("Input device created: %1").arg(device->deviceInfo().deviceName()));
}

void MainWindow::outputDeviceCreated(const QMidiOutput* device) {
    addLine(QString("Output device created: %1").arg(device->deviceInfo().deviceName()));
}

void MainWindow::inputDeviceDeleted(QMidiDeviceInfo info) {
    addLine(QString("Input device deleted: %1").arg(info.deviceName()));
}

void MainWindow::outputDeviceDeleted(QMidiDeviceInfo info) {
    addLine(QString("Output device deleted: %1").arg(info.deviceName()));
}

void MainWindow::receiveMessage(const QMidiMessage &m) {
    QString out(QString::fromLatin1("Midi message received. "));
    if (m.size() > 0) {
        out += QString("Cmd: %1").arg(m[0]);
    }
    for (int i = 1; i < m.size(); ++i) {
        out += QString(", Byte%1: %2").arg(QString::number(i), QString::number(m[i]));
    }
    addLine(out);
}

void MainWindow::inputIndexChanged(int index) {
    if (mCurrentInputDevice.isNull() == false) {
        QMidiInput::release(mCurrentInputDevice);
        mCurrentInputDevice = QMidiDeviceInfo();
    }

    if (index >= 0) {
        mCurrentInputDevice = mInputDevices->itemData(index).value<QMidiDeviceInfo>();
        if (mCurrentInputDevice.isNull()) {
            return;
        }
        const QMidiInput *input = QMidiInput::access(mCurrentInputDevice);
        if (input) {
            connect(input, &QMidiInput::notify, this, &MainWindow::receiveMessage);
            const QMidiOutput *output = QMidiOutput::access(mCurrentOutputDevice);
            if (output) {
                connect(input, &QMidiInput::notify, output, &QMidiOutput::receiveMidiMessage);
            }
        }
    }
}

void MainWindow::outputIndexChanged(int index) {
    if (mCurrentOutputDevice.isNull() == false) {
        QMidiOutput::release(mCurrentOutputDevice);
        mCurrentOutputDevice = QMidiDeviceInfo();
    }

    if (index >= 0) {
        mCurrentOutputDevice = mOutputDevices->itemData(index).value<QMidiDeviceInfo>();
        if (mCurrentOutputDevice.isNull()) {
            return;
        }
        const QMidiOutput *output = QMidiOutput::access(mCurrentOutputDevice);
        if (output) {
            const QMidiInput *input = QMidiInput::access(mCurrentInputDevice);
            if (input) {
                connect(input, &QMidiInput::notify, output, &QMidiOutput::receiveMidiMessage);
            }
        }
    }
}

