#include "mainwindow.h"
#include <QMidiInput>
#include <QMidiSystemNotifier>
#include <QDebug>

MainWindow::MainWindow(QWidget* parent)
    : QMainWindow(parent) {

    mConsoleOutput = new QPlainTextEdit(this);
    setCentralWidget(mConsoleOutput);

    mConsoleOutput->setReadOnly(true);
    QFont font("Courier New");
    mConsoleOutput->setFont(font);

    addLine("MIDI Device Monitor");
    addLine("===================");
    addLine("");

    auto notifier = new QMidiSystemNotifier(this);
    connect(notifier, &QMidiSystemNotifier::inputDeviceAttached, this, &MainWindow::inputDeviceAttached);
    connect(notifier, &QMidiSystemNotifier::outputDeviceAttached, this, &MainWindow::outputDeviceAttached);
    connect(notifier, &QMidiSystemNotifier::inputDeviceDetached, this, &MainWindow::inputDeviceDetached);
    connect(notifier, &QMidiSystemNotifier::outputDeviceDetached, this, &MainWindow::outputDeviceDetached);
    connect(notifier, &QMidiSystemNotifier::inputDeviceCreated, this, &MainWindow::inputDeviceCreated);

    printInputDevices();
    printOutputDevices();
    
    auto inDevice = QMidiInput::access();
    connect(inDevice, &QMidiInput::notify, this, &MainWindow::receiveMessage);
    
}

MainWindow::~MainWindow() {
}

void MainWindow::addLine(QString line) {
    //QTextCursor prev_cursor = mConsoleOutput->textCursor();
    mConsoleOutput->moveCursor(QTextCursor::End);
    mConsoleOutput->insertPlainText(line + "\n");
    mConsoleOutput->moveCursor(QTextCursor::End);
    //mConsoleOutput->setTextCursor(prev_cursor);
}

void MainWindow::printInputDevices() {
    addLine("Available input devices:");
    for (QMidiDeviceInfo &info : QMidiDeviceInfo::availableDevices(QMidi::MidiInput)) {
        addLine(QString("\t- %1").arg(info.deviceName()));
    }
    addLine(QString());
}

void MainWindow::printOutputDevices() {
    addLine("Available output devices:");
    for (QMidiDeviceInfo &info : QMidiDeviceInfo::availableDevices(QMidi::MidiOutput)) {
        addLine(QString("\t- %1").arg(info.deviceName()));
    }
    addLine(QString());
}

void MainWindow::inputDeviceAttached(const QMidiDeviceInfo info) {
    addLine(QString("Input device attached: %1").arg(info.deviceName()));
}

void MainWindow::outputDeviceAttached(const QMidiDeviceInfo info) {
    addLine(QString("Output device attached: %1").arg(info.deviceName()));
}

void MainWindow::inputDeviceDetached(const QMidiDeviceInfo info) {
    addLine(QString("Input device detached: %1").arg(info.deviceName()));
}

void MainWindow::outputDeviceDetached(const QMidiDeviceInfo info) {
    addLine(QString("Output device detached: %1").arg(info.deviceName()));
}

void MainWindow::inputDeviceCreated(const QMidiInput* device) {
    addLine(QString("Input device created: %1").arg(device->deviceInfo().deviceName()));
}

void MainWindow::receiveMessage(const QMidiMessage &m) {
    QString out(QString::fromLatin1("Midi message received. "));
    if (m.size() > 0) {
        out += QString("Cmd: %1").arg(m[0]);
    }
    for (int i = 1; i < m.size(); ++i) {
        out += QString(", Byte%1: %2").arg(QString::number(i), QString::number(m[i]));
    }
    addLine(out);
}

/*
void MainWindow::inputDeviceAttached(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Input device attached: %1").arg(QString::fromStdString(id->humanReadable()))));
    QMetaObject::invokeMethod(this, "printInputDevices");

}
void MainWindow::outputDeviceAttached(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Output device attached: %1").arg(QString::fromStdString(id->humanReadable()))));
    QMetaObject::invokeMethod(this, "printOutputDevices");
}

void MainWindow::inputDeviceDetached(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Input device detached: %1").arg(QString::fromStdString(id->humanReadable()))));
    QMetaObject::invokeMethod(this, "printInputDevices");
}
void MainWindow::outputDeviceDetached(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Output device detached: %1").arg(QString::fromStdString(id->humanReadable()))));
    QMetaObject::invokeMethod(this, "printOutputDevices");
}

void MainWindow::inputDeviceCreated(umidi::MidiInputDevicePtr device) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Input device created: %1").arg(QString::fromStdString(device->id()->humanReadable()))));
    device->addListener(this);
}
void MainWindow::outputDeviceCreated(umidi::MidiOutputDevicePtr device) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Output device created: %1").arg(QString::fromStdString(device->id()->humanReadable()))));
}

void MainWindow::inputDeviceDeleted(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Input device deleted: %1").arg(QString::fromStdString(id->humanReadable()))));
}
void MainWindow::outputDeviceDeleted(umidi::MidiDeviceID id) {
    QMetaObject::invokeMethod(this, "addLine", Q_ARG(QString, QString("Output device deleted: %1").arg(QString::fromStdString(id->humanReadable()))));
}

*/
