#include "qmidiautoconnector.h"
#include <QtCore/private/qobject_p.h>
#include <QtCore/qlist.h>
#include <QtCore/qbytearray.h>
#include <QtCore/qdatastream.h>
#include <QtCore/qdebug.h>
#include <QtMidi/qmidisystemnotifier.h>
#include <QDebug>

QT_BEGIN_NAMESPACE

class QMidiAutoConnectorPrivate {
public:
    QMidiAutoConnectorPrivate(QMidiAutoConnector *connector)
    {
        q = connector;
        mAutoConnect[QMidi::MidiInput] = true;
        mAutoConnect[QMidi::MidiOutput] = true;
        mForceSingleDevice[QMidi::MidiInput] = true;
        mForceSingleDevice[QMidi::MidiOutput] = true;

        mNotifier = new QMidiSystemNotifier();
    }

    virtual ~QMidiAutoConnectorPrivate()
    {
        for (auto mode : {QMidi::MidiInput, QMidi::MidiOutput}) {
            for (auto dev : mCurrentDevices[mode]) {
                delete dev;
            }
        }
        delete mNotifier;
    }

public:
    bool isAutoConnectEnabled(QMidi::Mode mode) const {return mAutoConnect[mode];}
    bool isSingleDeviceForced(QMidi::Mode mode) const {return mForceSingleDevice[mode];}

    void setAutoConnect(QMidi::Mode mode, bool autoconnect = true) {
        if (mAutoConnect[mode] != autoconnect) {
            mAutoConnect[mode] = autoconnect;
            emit q->autoConnectChanged(mode, autoconnect);
            if (mode == QMidi::MidiInput) {
                emit q->autoConnectToInputChanged(autoconnect);
            } else {
                emit q->autoConnectToOutputChanged(autoconnect);
            }

            if (autoconnect) {
                // if we are not connected to a device yet, we connect now
                if (mForceSingleDevice[mode]) {
                    // only single device => create default device if there is
                    // no device present
                    if (mCurrentDevices[mode].size() == 0) {
                        createDefaultDevice(mode);
                    }
                } else {
                    // connect to all devices
                    QList<QMidiDeviceInfo> infos = QMidiDeviceInfo::availableDevices(mode);
                    for (auto &info : infos) {
                        connectWith(info, mode);
                    }
                }
            }
        }
    }

    void setForceSingleDevice(QMidi::Mode mode, bool force = true) {
        if (mForceSingleDevice[mode] != force) {
            mForceSingleDevice[mode] = force;
            emit q->forceSingleDeviceChanged(mode, force);
            if (mode == QMidi::MidiInput) {
                emit q->forceSingleInputDeviceChanged(force);
            } else {
                emit q->forceSingleOutputDeviceChanged(force);
            }

            if (force) {
                // only keep last device
                while (mCurrentDevices[mode].size() > 1) {
                    deleteDevice(mCurrentDevices[mode].front(), mode);
                }
            } else {
                // connect to all devices if, auto connect is enabled
                if (mAutoConnect[mode]) {
                    QList<QMidiDeviceInfo> infos = QMidiDeviceInfo::availableDevices(mode);
                    for (auto &info : infos) {
                        connectWith(info, mode);
                    }
                }
            }
        }
    }

    void restore(const QByteArray &array, bool notify = true) {
        QDataStream d(array);
        bool buffer;
        if (notify) {
            for (auto i : {QMidi::MidiInput, QMidi::MidiOutput}) {
                d >> buffer;
                setAutoConnect(i, buffer);
                d >> buffer;
                setForceSingleDevice(i, buffer);
            }
        } else {
            for (auto i : {QMidi::MidiInput, QMidi::MidiOutput}) {
                d >> mAutoConnect[i] >> mForceSingleDevice[i];
            }
        }
    }

    QByteArray store() const {
        QByteArray data;
        QDataStream d(&data, QIODevice::WriteOnly);

        for (auto i : {QMidi::MidiInput, QMidi::MidiOutput}) {
            d << mAutoConnect[i] << mForceSingleDevice[i];
        }

        return data;
    }

    void connectWith(const QMidiDeviceInfo &info, QMidi::Mode mode)
    {
        for (QMidiDevice *dev : mCurrentDevices[mode]) {
            if (dev->deviceInfo() == info) {
                // already connected
                return;
            }
        }

        if (mForceSingleDevice[mode]) {
            deleteAllDevices(mode);
        }
        createDevice(info, mode);
    }

    void disconnectFrom(const QMidiDeviceInfo &info, QMidi::Mode mode)
    {
        for (auto dev : mCurrentDevices[mode]) {
            if (dev->deviceInfo() == info) {
                deleteDevice(dev, mode);
                break;
            }
        }

        if (mAutoConnect[mode]) {
            if (mCurrentDevices[mode].size() == 0) {
                // connect to default device
                createDefaultDevice(mode);
            }
        }
    }

    void disconnectFromAll(QMidi::Mode mode)
    {
        deleteAllDevices(mode);

        if (mAutoConnect[mode]) {
            createDefaultDevice(mode);
        }
    }

private:
    void onAttached(const QMidiDeviceInfo &info, QMidi::Mode mode)
    {
        if (mAutoConnect[mode]) {
            connectWith(info, mode);
        }
    }

    void onDetached(const QMidiDeviceInfo &info, QMidi::Mode mode)
    {
        disconnectFrom(info, mode);
    }

    void createDevice(const QMidiDeviceInfo &info, QMidi::Mode mode)
    {
        if (info.isNull()) {return;}

        if (mode == QMidi::MidiInput) {
            auto dev = new QMidiInput(info);
            mCurrentDevices[mode].push_back(dev);
            emit q->inputDeviceCreated(dev);
            emit q->deviceCreated(dev, mode);
        } else {
            auto dev = new QMidiOutput(info);
            mCurrentDevices[mode].push_back(dev);
            emit q->outputDeviceCreated(dev);
            emit q->deviceCreated(dev, mode);
        }
    }

    void createDefaultDevice(QMidi::Mode mode)
    {
        QMidiDeviceInfo info;
        if (mode == QMidi::MidiInput) {
            info = QMidiDeviceInfo::defaultInputDevice();
        } else {
            info = QMidiDeviceInfo::defaultOutputDevice();
        }

        createDevice(info, mode);
    }

    void deleteDevice(QMidiDevice *dev, QMidi::Mode mode)
    {
        const QMidiDeviceInfo info = dev->deviceInfo();
        mCurrentDevices[mode].removeOne(dev);
        dev->deleteLater();
        emit q->deviceDeleted(info, mode);
        if (mode == QMidi::MidiInput) {
            emit q->inputDeviceDeleted(info);
        } else {
            emit q->outputDeviceDeleted(info);
        }
    }

    void deleteAllDevices(QMidi::Mode mode) {
        while (mCurrentDevices[mode].size() > 0) {
            deleteDevice(mCurrentDevices[mode].first(), mode);
        }
    }

private:
    // always two for input and output
    bool mAutoConnect[2];
    bool mForceSingleDevice[2];
    QList<QMidiDevice*> mCurrentDevices[2];
    QMidiSystemNotifier *mNotifier;
    QMidiAutoConnector *q;

    friend class QMidiAutoConnector;
};

/*!
    \class QMidiAutoConnector
    \brief The QMidiAutoConnector class provides an optional tool to automatically connect to midi input or output device.

    \inmodule QtMidi
    \ingroup midi
    \ingroup midi_audio


    The QMidiAutoConnector can be used to automatically connect to available input and output midi devices.
    You can specify if there may only be one device of the given QMidi::Mode connected at the same time via
    \l setAllowMultiDevices.

    The following snippet shows how you can create a QMidiAutoConnector and listen to deviceCreated and deviceDeleted events.

    \snippet midi-snippets/midi.cpp Auto connector

    In you custom implementation you should listen to the deviceCreated signal and connect the \l QMidiInput::notify method to
    you \l QMidiMessage listener.

    The \l setForceSingleDevicem and \l setAllowMultiDevices are inverse functions.
 */

/*!
    Construct a QMidiAutoConnector and attach it to \a parent.
 */
QMidiAutoConnector::QMidiAutoConnector(QObject *parent)
    : QObject(parent)
{
    d = new QMidiAutoConnectorPrivate(this);

    connect(d->mNotifier, &QMidiSystemNotifier::inputDeviceAttached, [this](const QMidiDeviceInfo info) {
        d->onAttached(info, QMidi::MidiInput);
    });
    connect(d->mNotifier, &QMidiSystemNotifier::inputDeviceDetached, [this](const QMidiDeviceInfo info) {
        d->onDetached(info, QMidi::MidiInput);
    });
    connect(d->mNotifier, &QMidiSystemNotifier::outputDeviceAttached, [this](const QMidiDeviceInfo info) {
        d->onAttached(info, QMidi::MidiOutput);
    });
    connect(d->mNotifier, &QMidiSystemNotifier::outputDeviceDetached, [this](const QMidiDeviceInfo info) {
        d->onDetached(info, QMidi::MidiOutput);
    });
}

/*!
    Destroy this auto connector.
 */
QMidiAutoConnector::~QMidiAutoConnector()
{
    delete d;
}

/*!
    Returns if auto connect is enabled for the given \a mode.
 */
bool QMidiAutoConnector::isAutoConnectEnabled(QMidi::Mode mode) const
{
    return d->isAutoConnectEnabled(mode);
}

/*!
    Returns if only a single device may be connected at the same time for the given \a mode.

    \sa allowMultiDevices
 */
bool QMidiAutoConnector::isSingleDeviceForced(QMidi::Mode mode) const
{
    return d->isSingleDeviceForced(mode);
}

/*!
    Returns if the auto connector is allowed to connect to multiple devices at the same time
    for a given \a mode.

    \sa isSingleDeviceForced
 */
bool QMidiAutoConnector::allowMultiDevices(QMidi::Mode mode) const
{
    return !d->isSingleDeviceForced(mode);
}

/*!
    \fn QList<const DeviceType*> QMidiAutoConnector::devices() const
    Returns all of the current connected devices.

    \sa allowMultiDevices
 */
template <class DeviceType>
QList<const DeviceType*> QMidiAutoConnector::devices() const
{
    QList<const DeviceType*> devices;
    for (auto dev : d->mCurrentDevices[DeviceType::Mode]) {
        devices.push_back(static_cast<const DeviceType*>(dev));
    }
    return devices;
}

template Q_MIDI_EXPORT QList<const QMidiInput*> QMidiAutoConnector::devices() const;

template Q_MIDI_EXPORT QList<const QMidiOutput*> QMidiAutoConnector::devices() const;

/*!
    \fn const DeviceType* QMidiAutoConnector::device() const
    Returns the last connected device.

    \sa isSingleDeviceForced
 */
template <class DeviceType>
const DeviceType* QMidiAutoConnector::device() const {
    if (d->mCurrentDevices[DeviceType::Mode].size() > 0) {
        return static_cast<const DeviceType*>(d->mCurrentDevices[DeviceType::Mode].last());
    }
    return nullptr;
}

template Q_MIDI_EXPORT const QMidiInput* QMidiAutoConnector::device() const;

template Q_MIDI_EXPORT const QMidiOutput* QMidiAutoConnector::device() const;

/*!
    Restores the internal configuration from the given \a array.
    If \a notify is set to true signals for changed parameters will
    be emitted.

    Usually the \a array should be created by \l store.

    \sa store
 */
void QMidiAutoConnector::restore(const QByteArray &array, bool notify) {
    d->restore(array, notify);
}

/*!
    Returns a QByteArray representing the configuration of the auto connector.
    This can be used to store the current settings.

    \sa restore
 */
QByteArray QMidiAutoConnector::store() const {
    return d->store();
}

/*!
    Set whether to \a autoconnect to devices of \a mode.
    If \a autoconnect is set to true it will instantly check if it shall
    connect to new devices depending on the \l isAutoConnectEnabled state.

    \sa isAutoConnectEnabled, setAutoConnectToInput, setAutoConnectToOutput
 */
void QMidiAutoConnector::setAutoConnect(QMidi::Mode mode, bool autoconnect)
{
    d->setAutoConnect(mode, autoconnect);
}

/*!
    Set whether to \a autoconnect to input devices.

    \sa setAutoConnect
 */
void QMidiAutoConnector::setAutoConnectToInput(bool autoconnect)
{
    setAutoConnect(QMidi::MidiInput, autoconnect);
}

/*!
    Set whether to \a autoconnect to output devices.

    \sa setAutoConnect
 */
void QMidiAutoConnector::setAutoConnectToOutput(bool autoconnect)
{
    setAutoConnect(QMidi::MidiOutput, autoconnect);
}

/*!
    If \a force is true maximal one device of \a mode will be connected
    at the same time.

    If set to true it will disconnect all devices and connect to
    the \l{QMidiInput::QMidiInput()}{default input device}.
 */
void QMidiAutoConnector::setForceSingleDevice(QMidi::Mode mode, bool force)
{
    d->setForceSingleDevice(mode, force);
}

/*!
    If \a force is true maximal one input device will be connected
    at the same time.

    \sa setForceSingleDevice
 */
void QMidiAutoConnector::setForceSingleInputDevice(bool force)
{
    setForceSingleDevice(QMidi::MidiInput, force);
}
/*!
    If \a force is true maximal one input device will be connected
    at the same time.

    \sa setForceSingleDevice
 */
void QMidiAutoConnector::setForceSingleOutputDevice(bool force)
{
    setForceSingleDevice(QMidi::MidiOutput, force);
}

/*!
    If \a allow is set to true the auto connector is allowed to connect to
    multiple devices of type \a mode.
 */
void QMidiAutoConnector::setAllowMultiDevices(QMidi::Mode mode, bool allow)
{
    d->setForceSingleDevice(mode, !allow);
}

/*!
    If \a allow is set to true the auto connector is allowed to
    connect to multiple input devices.

    \sa setAllowMultiDevices
 */
void QMidiAutoConnector::setAllowMultiInputDevices(bool allow)
{
    setAllowMultiDevices(QMidi::MidiInput, allow);
}

/*!
    It \a allow is set to true the auto connector is allowed to
    connect to multiple output devices.

    \sa setAllowMultiDevices
 */
void QMidiAutoConnector::setAllowMultiOutputDevices(bool allow)
{
    setAllowMultiDevices(QMidi::MidiOutput, allow);
}

/*!
    Force the auto connector to connect with the device of type
    \a mode with \a info.

    If \l isSingleDeviceForced is set the auto connector will
    disconnect from the currrent device.
 */
void QMidiAutoConnector::connectWith(const QMidiDeviceInfo &info, QMidi::Mode mode)
{
    d->connectWith(info, mode);
}

/*!
    Force the auto connect to disconnect from the device of type
    \a mode with \a info.

    If \l isAutoConnectEnabled is true ts will connect to a new device
    instantly.
 */
void QMidiAutoConnector::disconnectFrom(const QMidiDeviceInfo &info, QMidi::Mode mode)
{
    d->disconnectFrom(info, mode);
}

/*!
    Force the auto connect to disconnect from all devices of type \a mode.

    If \l isAutoConnectEnabled is true it will connect to the default device
    instantly.
 */
void QMidiAutoConnector::disconnectFromAll(QMidi::Mode mode)
{
    d->disconnectFromAll(mode);
}

/*!
    \fn QMidiAutoConnector::autoConnectChanged(QMidi::Mode mode, bool autoconnect)
    This signal is emitted if the auto connect state of \a mode changed to \a autoconnect.

    \sa setAutoConnect, isAutoConnectEnabled
 */

/*!
    \fn QMidiAutoConnector::forceSingleDeviceChanged(QMidi::Mode mode, bool force)
    This signal is emitted if the state of forcing a single \a mode device changed to \a force.

    \sa setForceSingleDevice, isSingleDeviceForced
 */

/*!
    \fn QMidiAutoConnector::deviceCreated(const QMidiDevice* device, QMidi::Mode mode)
    This signal is emitted if a \a device of type \a mode was created by
    the auto connector.

    \sa inputDeviceCreated, outputDeviceCreated, deviceDeleted
 */

/*!
    \fn QMidiAutoConnector::inputDeviceCreated(const QMidiInput *device)
    This signal is emitted if a input \a device was created.

    \sa deviceCreated, outputDeviceCreated, inputDeviceDeleted
 */

/*!
    \fn QMidiAutoConnector::outputDeviceCreated(const QMidiOutput *device)
    This signal is emitted if a output \a device was created.

    \sa deviceCreated, inputDeviceCreated, outputDeviceDeleted
 */

/*!
    \fn QMidiAutoConnector::deviceDeleted(const QMidiDeviceInfo info, QMidi::Mode mode)
    This signal is emitted if a device with given \a info of \a mode was destroyed
    by the auto connector.

    \sa inputDeviceDeleted, outputDeviceDeleted, deviceCreated
 */

/*!
    \fn QMidiAutoConnector::inputDeviceDeleted(const QMidiDeviceInfo info)
    This signal is emitted if a input device with given \a \info was destroyed.

    \sa deviceDeleted, outputDeviceDeleted, inputDeviceCreated
 */

/*!
    \fn QMidiAutoConnector::outputDeviceDeleted(const QMidiDeviceInfo info)
    This signal is emitted if a output device with given \a info was destroyed.

    \sa deviceDeleted, outputDeviceDeleted, outputDeviceCreated
 */

QT_END_NAMESPACE
